%% Copyright (C) 2025 by sysmocom - s.f.m.c. GmbH <info@sysmocom.de>
%% Author: Vadim Yanitskiy <vyanitskiy@sysmocom.de>
%%
%% All Rights Reserved
%%
%% SPDX-License-Identifier: AGPL-3.0-or-later
%%
%% This program is free software; you can redistribute it and/or modify
%% it under the terms of the GNU Affero General Public License as
%% published by the Free Software Foundation; either version 3 of the
%% License, or (at your option) any later version.
%%
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%%
%% You should have received a copy of the GNU Affero General Public License
%% along with this program.  If not, see <https://www.gnu.org/licenses/>.
%%
%% Additional Permission under GNU AGPL version 3 section 7:
%%
%% If you modify this Program, or any covered work, by linking or
%% combining it with runtime libraries of Erlang/OTP as released by
%% Ericsson on https://www.erlang.org (or a modified version of these
%% libraries), containing parts covered by the terms of the Erlang Public
%% License (https://www.erlang.org/EPLICENSE), the licensors of this
%% Program grant you additional permission to convey the resulting work
%% without the need to license the runtime libraries of Erlang/OTP under
%% the GNU Affero General Public License. Corresponding Source for a
%% non-source form of such a combination shall include the source code
%% for the parts of the runtime libraries of Erlang/OTP used as well as
%% that of the covered work.

-module(logger_gsmtap_h).

-moduledoc """
GSMTAP logging handler.

This is a logging handler for Erlang's built-in Logger module.
It sends log messages via GSMTAP (udp/4729), allowing capture of
log messages alongside with other PDUs sent/received by the program.

GSMTAP [1] is a pseudo-header initially designed to encapsulate GSM air interface
(Um) frames within UDP/IP packets.  Over time, its functionality was extended
to enable the transport of logging messages.  Wireshark does support GSMTAP [2].

[1] https://osmocom.org/projects/baseband/wiki/GSMTAP
[2] https://wiki.wireshark.org/GSMTAP

To add a new instance of the GSMTAP handler, use
[`logger:add_handler/3`](`logger:add_handler/3`).  The handler configuration
argument is a map which can contain general configuration parameters, as
documented in the [`User's Guide`](logger_chapter.md#handler-configuration),
and handler specific parameters.  The specific data is stored in a sub map
with the key `config`, and can contain the following parameters:

- **`loc_addr = ` `t:string/0` ` | ` `t:inet:ip_address/0`**{: #loc_addr } -
  This specifies the local (bind) address to originate GSMTAP PDUs from.

  Defaults to `"0.0.0.0"`, letting the OS to select it automatically.

- **`rem_addr = ` `t:string/0` ` | ` `t:inet:ip_address/0`**{: #rem_addr } -
  This specifies the remote address to send GSMTAP PDUs to.

  Defaults to `"127.0.0.1"` (localhost).

- **`app_name = ` `t:string/0` ` | ` `t:atom/0`**{: #app_name } -
  This specifies the application name to be included in GSMTAP PDUs.

  Defaults to `unknown`.

Modification of these parameters is currently not supported.

Example of adding a GSMTAP handler:

```erlang
logger:add_handler(gsmtap, logger_gsmtap_h,
                   #{config => #{rem_addr => "127.0.0.1",
                                 app_name => "MyApp"}}).
```
""".

-behaviour(logger_handler).

%% logger_handler callbacks
-export([adding_handler/1,
         removing_handler/1,
         filter_config/1,
         log/2]).

-define(LADDR_DEFAULT, "0.0.0.0").
-define(RADDR_DEFAULT, "127.0.0.1").
-define(APP_NAME_DEFAULT, unknown).


%% ------------------------------------------------------------------
%% logger_handler API
%% ------------------------------------------------------------------

-spec adding_handler(logger_handler:config()) -> {ok, logger_handler:config()} |
                                                 {error, term()}.
adding_handler(Config) ->
    HConfig = maps:get(config, Config, #{}),
    LAddr = ipaddr(maps:get(loc_addr, HConfig, ?LADDR_DEFAULT)),
    RAddr = ipaddr(maps:get(rem_addr, HConfig, ?RADDR_DEFAULT)),
    AppName = app_name(maps:get(app_name, HConfig, ?APP_NAME_DEFAULT)),
    {ok, Pid} = logger_gsmtap:start(LAddr, RAddr, AppName),
    {ok, Config#{config => HConfig#{pid => Pid,
                                    loc_addr => LAddr,
                                    rem_addr => RAddr,
                                    app_name => AppName}}}.


-spec removing_handler(logger_handler:config()) -> ok.
removing_handler(#{config := #{pid := Pid}}) ->
    logger_gsmtap:stop(Pid),
    ok.


-spec filter_config(logger_handler:config()) -> logger_handler:config().
filter_config(#{config := HConfig} = Config) ->
    Config#{config => maps:without([pid], HConfig)}.


-spec log(logger:log_event(), logger_handler:config()) -> term().
log(LogEvent, #{config := #{pid := Pid}}) ->
    logger_gsmtap:log(Pid, LogEvent).


%% ------------------------------------------------------------------
%% private API
%% ------------------------------------------------------------------

-spec ipaddr(Addr0) -> Addr1
    when Addr0 :: string() | inet:ip_address(),
         Addr1 :: inet:ip_address().
ipaddr(AddrStr) when is_list(AddrStr) ->
    {ok, Addr} = inet:parse_address(AddrStr),
    Addr;

ipaddr(Addr) ->
    true = inet:is_ip_address(Addr),
    Addr.


-spec app_name(string() | atom()) -> string().
app_name(AppName) when is_list(AppName) ->
    AppName;

app_name(AppName) when is_atom(AppName) ->
    atom_to_list(AppName).


%% vim:set ts=4 sw=4 et:
